#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include "hash.h"

// Cantor pairing function
int cantor(int x, int y) {
    return ((x + y) * (x + y + 1) / 2) + y;
}

int generate_hash(int x, int y) {
    if (x < 0) {
        x = -2*x - 1;
    } else {
        x = 2 * x;
    }
    if (y < 0) {
        y = -2*y - 1;
    } else {
        y = 2 * y;
    }
    return cantor(x, y);
}

void freeGrid(Grid grid) {
    for (int i = 0; i < grid->capacity; i++) {
        List list = grid->data[i];
        while (list != NULL) {
            List next = list->next;
            free(list);
            list = next;
        }
    }
    free(grid->data);
    free(grid);
}

void subGridInsert(Grid grid, int hash, chunk* elt, int x, int y) {
    int index = hash % grid->capacity;
    //printf("               inserting at index [%d]", index);
    List* data = grid->data;
    if (data[index] == NULL) {
        //printf("                   <-->\n");
        List list = malloc(sizeof(struct List));
        list->hash = hash;
        list->x = x;
        list->y = y;
        list->elt = elt;
        list->next = NULL;
        grid->size++;
        data[index] = list;
    } else {
        //printf("                   <>\n");
        List list = data[index];
        while (list->next != NULL && list->hash != hash) {
            list = list->next;
        };
        //printf("                   <>\n");
        if (list->x == x && list->y == y) {
            list->elt = elt;
        } else {
            List nex = malloc(sizeof(struct List));
            nex->hash = hash;
            nex->x = x;
            nex->y = y;
            nex->elt = elt;
            nex->next = NULL;
            list->next = nex;
            grid->size++;
        }
    }
}

void gridInsert(Grid grid, int x, int z, chunk* elt) {
    if (grid->size >= grid->capacity * grid->loadFactor) {
        grid->size = 0;
        grid->capacity *= 2;

        printf("resizing\n");

        List* old_data = grid->data;
        grid->data = malloc(sizeof(List) * grid->capacity);;
        for (int i = 0; i < grid->capacity; i++) {
            grid->data[i] = NULL;
        };
        for (int i = 0; i < grid->capacity/2; i++) {
            List list = old_data[i];
            while (list != NULL) {
                subGridInsert(grid, list->hash, list->elt, list->x, list->y);
                list = list->next;
            }
        }
    };

    int hash = generate_hash(x, z);
    subGridInsert(grid, hash, elt, x, z);
}

chunk* gridGet(Grid grid, int x, int z) {
    int hash = generate_hash(x, z);
    int index = hash % grid->capacity;
    List list = grid->data[index];
    while (list != NULL && (list->x != x || list->y != z)) {
        list = list->next;
    }
    if (list == NULL) {
        return NULL;
    }
    return list->elt;
}

bool gridMem(Grid grid, int x, int z) {
    int hash = generate_hash(x, z);
    int index = hash % grid->capacity;
    List list = grid->data[index];
    while (list != NULL && (list->x != x || list->y != z)) {
        list = list->next;
    }
    if (list == NULL) {
        return false;
    }
    return true;
}

chunk* gridRemove(Grid grid, int x, int z) {
    int hash = generate_hash(x, z);
    int index = hash % grid->capacity;
    List list = grid->data[index];
    List prev = NULL;
    while (list != NULL && list->hash != hash) {
        prev = list;
        list = list->next;
    }
    if (list == NULL) {
        return NULL;
    }
    chunk* elt = list->elt;
    if (prev == NULL) {
        grid->data[index] = list->next;
    } else {
        prev->next = list->next;
    }
    free(list);
    grid->size--;
    return elt;
}

Grid newGrid() {
    Grid grid = malloc(sizeof(struct Grid));
    grid->data = malloc(sizeof(List) * 800);
    grid->size = 0;
    grid->capacity = 800;
    grid->loadFactor = 0.75;
    for (int i = 0; i < grid->capacity; i++) {
        grid->data[i] = NULL;
    }
    grid->push = &gridInsert;
    grid->pop = &gridRemove;
    grid->get = &gridGet;
    grid->mem = &gridMem;
    return grid;
}
 
chunk** gridIter(Grid grid) {
    chunk** array = malloc(sizeof(chunk*) * grid->size);
    int index = 0;
    for (int i = 0; i < grid->capacity; i++) {
        List list = grid->data[i];
        while (list != NULL) {
            array[index++] = list->elt;
            list = list->next;
        }
    }
    return array;
}

ChunkCoord newChunkCoord(int x, int z) {
    ChunkCoord coord = malloc(sizeof(struct ChunkCoord));
    coord->x = x;
    coord->z = z;
    return coord;
}

void freeChunkCoord(ChunkCoord coord) {
    free(coord);
}
